////////////////////////////////////////////////////////////////////////////////
// Parameters restriction:
// 1) CACHE_DELTA must be even as we unroll two iterations
// 2) NUM_STEPS and CACHE_STEP must be multiples of CACHE_DELTA.
//    Though current implemenation can be modified to accept
//    arbitrary cache parameters, "bad" (not multiples of warp size)
//    CACHE_STEP and NUM_STEPS values can break alignment constraints of
//    memory coalescing, thus hitting overall memory performance
////////////////////////////////////////////////////////////////////////////////
#define  CACHE_STEP 960
#define CACHE_DELTA 32
#define  CACHE_SIZE (CACHE_STEP + CACHE_DELTA)



//Number of time steps
#define NUM_STEPS 2048

//End prices storage for each thread block
//(NUM_STEPS + 1) node count is aligned to a multiple of 16 for coalescing
static __device__ double steps[BLOCK_N * (NUM_STEPS + 16)];


////////////////////////////////////////////////////////////////////////////////
// Process an array of OptN options on GPU
////////////////////////////////////////////////////////////////////////////////
__global__ void binomialOptionsGPU(
    double *d_CallResult,
    double *d_StockPrice,
    double *d_OptionStrike,
    double *d_OptionYears,
    double Riskfree,
    double Volatility,
    int OptN
){
    //Prices cache
    __shared__ double dataA[CACHE_SIZE];
    __shared__ double dataB[CACHE_SIZE];
    //End prices array for current thread block
    double *p = &steps[blockIdx.x * (NUM_STEPS + 16)];

    for(int opt = blockIdx.x; opt < OptN; opt += gridDim.x){
        const double S = d_StockPrice[opt];
        const double X = d_OptionStrike[opt];
        const double T = d_OptionYears[opt];
        
        //Time per step
        const double dT    = T * (1.0f / NUM_STEPS);
        const double VsdT  = Volatility * sqrt(dT);
        const double RdT   = Riskfree * dT;
        //Per-step cont. compounded riskless rate
        const double R     = exp(RdT);
        const double Rinv  = 1.0f / R;
        //Up move corresponding to given variance
        const double u     = exp(VsdT);
        //Corresponding down move
        const double d     = 1.0f / u;
        //Pseudo-probability of upward move
        const double Pu    = (R - d) / (u - d);
        //Pseudo-probability of downard move
        const double Pd    = 1.0f - Pu;
        //Compounded quotents
        const double PuByR = Pu * Rinv;
        const double PdByR = Pd * Rinv;


        ///////////////////////////////////////////////////////////////////////
        // Compute values at expiration date:
        // call option value at period end is V(T) = S(T) - X
        // if S(T) is greater than X, or zero otherwise.
        // The computation is similar for put options.
        ///////////////////////////////////////////////////////////////////////
        for(int i = threadIdx.x; i <= NUM_STEPS; i += blockDim.x){
            double price = S * expf(VsdT * (2.0f * i - NUM_STEPS));
            p[i]        = fmaxf(price - X, 0);
        }

        ////////////////////////////////////////////////////////////////////////
        // Walk backwards up binomial tree.
        // Can't do in-place reduction, since warps are scheduled out of order.
        // So double-buffer and synchronize to avoid read-after-write hazards.
        ////////////////////////////////////////////////////////////////////////
        for(int i = NUM_STEPS; i > 0; i -= CACHE_DELTA)
            for(int c_base = 0; c_base < i; c_base += CACHE_STEP){
                //Start and end positions within shared memory cache
                int c_start = min(CACHE_SIZE - 1, i - c_base);
                int c_end   = c_start - CACHE_DELTA;

                //Read data(with apron) to shared memory cache
                __syncthreads();
                for(int k = threadIdx.x; k <= c_start; k += blockDim.x)
                    dataA[k] = p[c_base + k];

                //Calculations within shared memory
                for(int k = c_start - 1; k >= c_end;){
                    //Compute discounted expected value
                    __syncthreads();
                    for(int l = threadIdx.x; l <= k; l += blockDim.x)
                        dataB[l] = PuByR * dataA[l + 1] + PdByR * dataA[l];
                    k--;
           
                    //Compute discounted expected value
                    __syncthreads();
                    for(int l = threadIdx.x; l <= k; l += blockDim.x)
                        dataA[l] = PuByR * dataB[l + 1] + PdByR * dataB[l];
                    k--;
                }

                //Flush shared memory cache
                __syncthreads();
                for(int k = threadIdx.x; k <= c_end; k += blockDim.x)
                    p[c_base + k] = dataA[k];
            }

        //Write the value at the top of the tree to destination buffer
        if(threadIdx.x == 0)
        {
           d_CallResult[opt] = dataA[0];
        }
    }
}

